@@ -13,11 +13,15 @@
 #include <unistd.h>
 #include "../lib/libunshield.h"
 #ifdef HAVE_CONFIG_H
-#include "../lib/unshield_config.h"
+#include "lib/unshield_config.h"
 #endif
 #if HAVE_FNMATCH
 #include <fnmatch.h>
 #endif
+#ifdef HAVE_ICONV
+#include <iconv.h>
+#include <errno.h>
+#endif
 
 #ifndef VERSION
 #define VERSION "Unknown"
@@ -25,6 +29,16 @@
 
 #define FREE(ptr)       { if (ptr) { free(ptr); ptr = NULL; } }
 
+#ifdef _WIN32
+  #define realpath(N,R) _fullpath((R),(N),_MAX_PATH)
+  #include <direct.h>
+  #ifndef PATH_MAX
+    #define PATH_MAX _MAX_PATH
+  #endif
+#else
+  #include <limits.h>
+#endif
+
 typedef enum 
 {
   OVERWRITE_ASK,
@@ -55,6 +69,7 @@ static const char* file_group_name    = NULL;
 static const char* component_name     = NULL;
 static bool junk_paths                = false;
 static bool make_lowercase            = false;
+static bool raw_filename              = false;
 static bool verbose                   = false;
 static ACTION action                  = ACTION_EXTRACT;
 static OVERWRITE overwrite            = OVERWRITE_ASK;
@@ -65,6 +80,10 @@ static int is_version                 = -1;
 static const char* cab_file_name      = NULL;
 static char* const* path_names        = NULL;
 static int path_name_count            = 0;
+#ifdef HAVE_ICONV
+static const char* encoding           = NULL;
+iconv_t encoding_descriptor           = (iconv_t)-1;
+#endif
 
 static bool make_sure_directory_exists(const char* directory)/*{{{*/
 {
@@ -91,7 +110,11 @@ static bool make_sure_directory_exists(const char* directory)/*{{{*/
 
       if (stat(current, &dir_stat) < 0)
       {
+        #ifdef __MINGW32__
+        if (_mkdir(current) < 0)
+        #else
         if (mkdir(current, 0700) < 0)
+        #endif
         {
           fprintf(stderr, "Failed to create directory %s\n", directory);
           goto exit;
@@ -110,12 +133,48 @@ exit:
   return success;
 }/*}}}*/
 
+#ifdef HAVE_ICONV
+static bool convert_encoding(char *buffer, size_t size)
+{
+  bool success = false;
+  char *newbuf, *inbuf, *outbuf;
+  size_t inbytesleft, outbytesleft, newsize;
+
+  if (encoding_descriptor == (iconv_t)-1)
+    return true;
+
+  inbuf = buffer;
+  inbytesleft = strlen(buffer);
+  newbuf = outbuf = malloc(size);
+  outbytesleft = size - 1;
+
+  if (iconv(encoding_descriptor,
+          &inbuf, &inbytesleft,
+          &outbuf, &outbytesleft) == (size_t)-1)
+  {
+    fprintf(stderr, "Could not encode text to '%s' error %s\n",
+        encoding, strerror(errno));
+    goto exit;
+  }
+
+  newsize = (size_t)(outbuf - newbuf);
+  memcpy(buffer, newbuf, newsize);
+  buffer[newsize] = '\0';
+
+  success = true;
+
+exit:
+  free(newbuf);
+  return success;
+}
+#endif
+
 static void show_usage(const char* name)
 {
   fprintf(stderr,
       "Syntax:\n"
       "\n"
-      "\t%s [-c COMPONENT] [-d DIRECTORY] [-D LEVEL] [-g GROUP] [-i VERSION] [-GhlOrV] c|g|l|t|x CABFILE [FILENAME...]\n"
+      "\t%s [-c COMPONENT] [-d DIRECTORY] [-D LEVEL] [-g GROUP] [-i VERSION] [-e ENCODING] [-GhlOrV] c|g|l|t|x CABFILE [FILENAME...]\n"
       "\n"
       "Options:\n"
       "\t-c COMPONENT  Only list/extract this component\n"
@@ -128,10 +187,12 @@ static void show_usage(const char* name)
       "\t-g GROUP      Only list/extract this file group\n"
       "\t-h            Show this help message\n"
       "\t-i VERSION    Force InstallShield version number (don't autodetect)\n"
+      "\t-e ENCODING   Convert filename character encoding to local codepage from ENCODING (implicitly sets -R)\n"
       "\t-j            Junk paths (do not make directories)\n"
       "\t-L            Make file and directory names lowercase\n"
       "\t-O            Use old compression\n"
       "\t-r            Save raw data (do not decompress)\n"
+      "\t-R            Don't do any conversion to file and directory names when extracting.\n"
       "\t-V            Print copyright and version information\n"
       "\n"
       "Commands:\n"
@@ -164,7 +225,7 @@ static bool handle_parameters(
 {
 	int c;
 
-	while ((c = getopt(argc, argv, "c:d:D:g:hi:jLnoOrV")) != -1)
+	while ((c = getopt(argc, argv, "c:d:D:g:hi:e:jLnoOrRV")) != -1)
 	{
 		switch (c)
     {
@@ -188,6 +249,16 @@ static bool handle_parameters(
         is_version = atoi(optarg);
         break;
 
+      case 'e':
+#ifdef HAVE_ICONV
+        encoding = optarg;
+        raw_filename = true;
+#else
+        fprintf(stderr, "This version of Unshield is not built with encoding support.\n");
+        return false;
+#endif
+        break;
+
       case 'j':
         junk_paths = true;
         break;
@@ -195,6 +266,10 @@ static bool handle_parameters(
       case 'L':
         make_lowercase = true;
         break;
+        
+      case 'R':
+        raw_filename = true;
+        break;
 
       case 'n':
         overwrite = OVERWRITE_NEVER;
@@ -288,6 +363,26 @@ static bool extract_file(Unshield* unshield, const char* prefix, int index)
   char filename[256];
   char* p;
   int directory = unshield_file_directory(unshield, index);
+  long int path_max;
+  char* real_output_directory;
+  char* real_filename;
+
+  #ifdef PATH_MAX
+    path_max = PATH_MAX;
+  #else
+    path_max = pathconf(path, _PC_PATH_MAX);
+    if (path_max <= 0)
+      path_max = 4096;
+  #endif
+
+  real_output_directory = malloc(path_max);
+  real_filename = malloc(path_max);
+  if (real_output_directory == NULL || real_filename == NULL)
+  {
+    fprintf(stderr,"Unable to allocate memory.");
+    success=false;
+    goto exit;
+  }
 
   strcpy(dirname, output_directory);
   strcat(dirname, "/");
@@ -325,14 +420,26 @@ static bool extract_file(Unshield* unshield, const char* prefix, int index)
         break;
 
       default:
-        if (!isprint(*p))
-          *p = '_';
-        else if (make_lowercase)
-          *p = tolower(*p);
+        if (!raw_filename)
+        {  
+          if (!isprint(*p))
+            *p = '_';
+          else if (make_lowercase)
+            *p = tolower(*p);
+        }
         break;;
     }
   }
 
+#ifdef HAVE_ICONV
+  if (!convert_encoding(dirname, sizeof(dirname)))
+  {
+    success = false;
+    goto exit;
+  }
+#endif
+
+
 #if 0
   if (dirname[strlen(dirname)-1] != '/')
     strcat(dirname, "/");
@@ -345,10 +452,39 @@ static bool extract_file(Unshield* unshield, const char* prefix, int index)
 
   for (p = filename + strlen(dirname); *p != '\0'; p++)
   {
-    if (!isprint(*p))
-      *p = '_';
-    else if (make_lowercase)
-      *p = tolower(*p);
+    if (!raw_filename)
+    {
+      if (!isprint(*p))
+        *p = '_';
+      else if (make_lowercase)
+        *p = tolower(*p);
+    }
+  }
+
+#ifdef HAVE_ICONV
+  if (!convert_encoding(filename + strlen(dirname),
+      sizeof(filename) - strlen(dirname)))
+  {
+    success = false;
+    goto exit;
+  }
+#endif
+
+  /* use GNU extension to return non-existing files to real_output_directory */
+  realpath(output_directory, real_output_directory);
+  realpath(filename, real_filename);
+  if (real_filename == NULL || strncmp(real_filename,
+                                       real_output_directory,
+                                       strlen(real_output_directory)) != 0)
+  {
+    fprintf(stderr, "\n\nExtraction failed.\n");
+    fprintf(stderr, "Possible directory traversal attack for: %s\n", filename);
+    fprintf(stderr, "To be placed at: %s\n\n", real_filename);
+    exit_status = 1;
+    success = false;
+    free(real_filename);
+    free(real_output_directory);
+    return success;
   }
 
   printf("  extracting: %s\n", filename);
@@ -365,6 +501,7 @@ static bool extract_file(Unshield* unshield, const char* prefix, int index)
       break;
   }
 
+exit:
   if (!success)
   {
     fprintf(stderr, "Failed to extract file '%s'.%s\n", 
@@ -373,7 +510,8 @@ static bool extract_file(Unshield* unshield, const char* prefix, int index)
     unlink(filename);
     exit_status = 1;
   }
-
+  free(real_filename);
+  free(real_output_directory);
   return success;
 }
 
@@ -578,6 +716,18 @@ int main(int argc, char* const argv[])
     goto exit;
   }
 
+#ifdef HAVE_ICONV
+  if (!unshield_is_unicode(unshield) && encoding != NULL)
+  {
+    if ((encoding_descriptor = iconv_open("", encoding)) == (iconv_t)-1)
+    {
+      fprintf(stderr, "Cannot use encoding '%s' error %s\n",
+          encoding, strerror(errno));
+      goto exit;
+    }
+  }
+#endif
+
   printf("Cabinet: %s\n", cab_file_name);
 
   switch (action)
@@ -609,6 +759,10 @@ int main(int argc, char* const argv[])
 
 exit:
   unshield_close(unshield);
+#ifdef HAVE_ICONV
+  if (encoding_descriptor != (iconv_t)-1)
+    iconv_close(encoding_descriptor);
+#endif
   if (!success)
     exit_status = 1;
   return exit_status;
